from read_line import Read_line
from sensor_drv import SensorDrv
import config
import obs
import omc048
import uasyncio
from util import lookup_parameters

class MantaResponse():
    def __init__(self, command):
        self.command = command
        self.value = b''

class Manta(SensorDrv):
    __syslog = obs.system_log('MANTA')
    def __init__(self, serial, conf):
        SensorDrv.__init__(self)
        self.__available_parameters =  {
                                            "Temp_deg_C":       ("Temp C"             ,"C"            ,"TMPC"     ),
                                            "Temp_deg_F":       ("Temp F"             ,"F"            ,"TMPF"     ), 
                                            "SpCond_uS/cm":     ("Cond uS/cm"         ,"uS/cm"        ,"CONu"     ), 
                                            "SpCond_mS/cm":     ("Cond mS/cm"         ,"mS/cm"        ,"CONm"     ), 
                                            "Salinity_PSU":     ("Cond psu"           ,"PSU"          ,"CONp"     ), 
                                            "TDS_mg/l":         ("TDS"                ,"mg/L"         ,"TDSm"     ), 
                                            "pH_units":         ("pH"                 ,"pH"           ,"PH"       ), 
                                            "pH_mV":            ("pH mV"              ,"mV"           ,"PHv"      ), 
                                            "pH2_units":        ("pH2"                ,"pH"           ,"PH2"      ), 
                                            "pH2_mV":           ("pH2 mV"             ,"mV"           ,"PHv2"     ), 
                                            "pH3_units":        ("pH3"                ,"pH"           ,"PH3"      ), 
                                            "pH3_mV":           ("pH3 mV"             ,"mV"           ,"PHv3"     ),
                                            "ORP_mV":           ("ORP mV"             ,"mV"           ,"ORPv"     ),
                                            "HDO_%Sat":         ("HDO %Sat"           ,"%"            ,"HDO"      ),
                                            "HDO_mg/l":         ("HDO mg/l"           ,"mg/l"         ,"HDOm"     ),
                                            "DO_mg/l":          ("DO mg/l"            ,"mg/l"         ,"DOm"      ),
                                            "DO_%Sat":          ("DO %Sat"            ,"%"            ,"DO"       ),
                                            "ODO_mg/l":         ("ODO mg/l"           ,"mg/l"         ,"ODOm"     ),
                                            "ODO_%Sat":         ("ODO %Sat"           ,"%"            ,"ODO"      ),
                                            "Turb_FNU":         ("Turb FNU"           ,"FNU"          ,"TURf"     ),
                                            "Turb_NTU":         ("Turb NTU"           ,"NTU"          ,"TURn"     ),
                                            "Turb2_FNU":        ("Turb2 FNU"          ,"FNU"          ,"TURf2"    ),
                                            "TurbDig_NTU":      ("TurbDig NTU"        ,"NTU"          ,"TURn2"    ),
                                            "Depth_m":          ("Depth m"            ,"m"            ,"DEPm"     ),
                                            "Depth_ft":         ("Depth ft"           ,"ft"           ,"DEPf"     ),
                                            "Depth_psig":       ("Depth psig"         ,"psig"         ,"DEPp"     ),
                                            "BP_mmHg":          ("BP mmHg"            ,"mmHg"         ,"BP"       ),
                                            "Int_Batt_V":       ("Int Batt V"         ,"V"            ,"BAT"      ),
                                            "CablePower_V":     ("CablePower V"       ,"V"            ,"CBLP"     ),
                                            "Cl_mg/L":          ("Cl mg/L"            ,"mg/L"         ,"CLm"      ),
                                            "Cl_mV":            ("Cl mV"              ,"mV"           ,"CLv"      ),
                                            "NH4_mg/L-N":       ("NH4 mg/L-N"         ,"mg/L-N"       ,"NH4m"     ),
                                            "NH4_mV":           ("NH4 mV"             ,"mV"           ,"NH4v"     ),
                                            "NH3gas_mg/L-N":    ("NH3gas mg/L-N"      ,"mg/L-N"       ,"NH3m"     ),
                                            "NH3T_mg/L-N":      ("NH3T mg/L-N"        ,"mg/L-N"       ,"NH3t"     ),
                                            "NO3_mg/L-N":       ("NO3 mg/L-N"         ,"mg/L-N"       ,"NO3m"     ),
                                            "NO3_mV":           ("NO3 mV"             ,"mV"           ,"NO3v"     ),
                                            "Na_mg/L":          ("Na mg/L"            ,"mg/L"         ,"NAm"      ),
                                            "Na_mV":            ("Na mV"              ,"mV"           ,"NAv"      ),
                                            "Br_mg/L":          ("Br mg/L"            ,"mg/L"         ,"BRm"      ),
                                            "Br_mV":            ("Br mV"              ,"mV"           ,"BRv"      ),
                                            "Ca_mg/L":          ("Ca mg/L"            ,"mg/L"         ,"CAm"      ),
                                            "Ca_mV":            ("Ca mV"              ,"mV"           ,"CAv"      ),
                                            "eCO2_ppm":         ("eCO2 ppm"           ,"ppm"          ,"CO2p"     ),
                                            "eCO2_mV":          ("eCO2 mV"            ,"mV"           ,"CO2m"     ),
                                            "CO2_mV":           ("CO2 mV"             ,"mV"           ,"CO2v"     ),
                                            "Chl_ug/l":         ("Chl ug/l"           ,"ug/l"         ,"CHLu"     ),
                                            "Chl_red_ug/l":     ("Chl red_ug/l"       ,"ug/l"         ,"CHLr"     ),
                                            "bg_ppb":           ("bg ppb"             ,"ppb"          ,"BG"       ),
                                            "bgm_ppb":          ("bgm ppb"            ,"ppb"          ,"BGM"      ),
                                            "Rhod_ppb":         ("Rhod ppb"           ,"ppb"          ,"RHOp"     ),
                                            "CDOM_ppb":         ("CDOM ppb"           ,"ppb"          ,"CDOMp"    ),
                                            "CDOM2_ppb":        ("CDOM2 ppb"          ,"ppb"          ,"CDOMp2"   ),
                                            "Oil-Crude_ppb":    ("Oil-Crude ppb"      ,"ppb"          ,"OILc"     ),
                                            "Fluor_ppb":        ("Fluor ppb"          ,"ppb"          ,"FLR"      ),
                                            "Oil-Fine_ppb":     ("Oil-Fine ppb"       ,"ppb"          ,"OILf"     ),
                                            "TRYP_ppb":         ("TRYP ppb"           ,"ppb"          ,"TRYP"     ),
                                            "OB_ppb":           ("OB ppb"             ,"ppb"          ,"OB"       ),
                                            "EOSIN_ppb":        ("EOSIN ppb"          ,"ppb"          ,"EOSIN"    ),
                                            "FDOM_ppb":         ("FDOM ppb"           ,"ppb"          ,"FDOM"     ),
                                            "PTSA_ppb":         ("PTSA ppb"           ,"ppb"          ,"PTSA"     ),
                                            "%Transmission":    ("%Transmission"      ,"%"            ,"TRNS"     ),
                                            "Circ_Status":      ("Circ Status"        ,""             ,"CIRC"     ),
                                            "Clean_Status":     ("Clean Status"       ,""             ,"CLN"      ),
                                            "VentedDepth_m":    ("VentedDepth m"      ,"m"            ,"VDEPm"    ),
                                            "VentedDepth_ft":   ("VentedDepth ft"     ,"ft"           ,"VDEPf"    ),
                                            "VentedDepth_psig": ("VentedDepth psig"   ,"psig"         ,"VDEPp"    ),
                                            "Stage_m":          ("Stage m"            ,"m"            ,"STGm"     ),
                                            "Stage_ft":         ("Stage ft"           ,"ft"           ,"STGf"     ),
                                            "Stage_psig":       ("Stage psig"         ,"psig"         ,"STGp"     ),
                                            "TDG_mmHg":         ("TDG mmHg"           ,"mmHg"         ,"TDG"      ),
                                            "PAR_uMol":         ("PAR uMol"           ,"uMol"         ,"PARu"     ),
                                            "PAR2_uMol":        ("PAR2 uMol"          ,"uMol"         ,"PARu2"    ),
                                            "eTemp_deg_C":      ("eTemp deg C"        ,"C"            ,"ETMPC"    ),
                                            "eTemp_deg_F":      ("eTemp deg F"        ,"F"            ,"ETMPF"    ),
                                            "UV_Status":        ("UV_Status"          ,""             ,"UVSTAT"   ),
                                            "Therm1_deg_C":     ("Therm1 deg C"       ,"C"            ,"THRMC1"   ),
                                            "Therm1_deg_F":     ("Therm1 deg F"       ,"F"            ,"THRMF1"   ),
                                            "Therm2_deg_C":     ("Therm2 deg C"       ,"C"            ,"THRMC2"   ),
                                            "Therm2_deg_F":     ("Therm2 deg F"       ,"F"            ,"THRMF2"   ),
                                            "Turb_Mod":         ("Turb Mod"           ,""             ,"TURBm"    )}

        self.startup_time = 30
        self.__supply_port_on = config.get_config_params(conf, 'supply_port_always_on')
        self.__wiper_interval = config.get_config_or_default(conf, 'wiper_interval', False)
        if self.__wiper_interval and self.__supply_port_on == True:
            self.__wiper_cron = obs.cron(self.__wiper_interval)
            self.__syslog.info('Wiper interval set to: {}'.format(self.__wiper_interval))
            self.__wipe_initiator = 0
        self.__perform_wipe = False
        self.__delay_after_wipe = config.get_config_or_default(conf, 'delay_after_wipe', 0)
        self.__serial = serial
        self.__serial_data = Read_line(serial, b'\n')
 
    async def read_values(self):
        if self.__wiper_interval and self.__supply_port_on == True:
            now = omc048.RTC().datetime()
            x = self.__wiper_cron.next(now)
            if (obs.date_time_delta(now, x) > self.__wipe_initiator):
                self.__perform_wipe = True
            self.__wipe_initiator = obs.date_time_delta(now, self.__wiper_cron.next(now))

        response = await self.__get_sensor_response('gui -on')
        if (response.value == 'Gui mode on') or (response.value == '> Gui mode on'):
            if self.__perform_wipe == True:
                self.__syslog.info('Wipe command initiated')
                self.__send_command('wipe')
                #Note the manta does not respons with a wipe time estimation, a static time of 10 seconds is used
                self.__perform_wipe = False
                await uasyncio.sleep_ms(10000)
                if self.__delay_after_wipe != 0:
                    self.__syslog.info('Measurement delay after wipe for {} seconds started'.format(self.__delay_after_wipe))
                    await uasyncio.sleep_ms(self.__delay_after_wipe*1000)
            response = await self.__get_sensor_response('read')
            if response: 
                result_split = response.value.split(',')
                #Note the result_split is parsed without the first three values, these contain 'date', 'time' and 'unknown'
                self.data = result_split[3:]
                return True

    async def detect_parameters(self, tag_number = ''):
        await self.flush()
        response = await self.__get_sensor_response('TO 0')
        if (response.value == 'Timeout time 0') or (response.value == '> Timeout time 0'):
            #In case the sensors comes out of sleep or was already powered a '> ' character is left in the uart buffer
            #which must be flushed out prior to obtaining the paramters
            await self.flush()
            response = await self.__get_sensor_response('readparameters')
            if response:
                try:
                    result_split = response.value.split(',')
                    #Note the result_split is parsed whithout the last value, this is an empty field due to a comma in the sensors response end
                    self.parameters = lookup_parameters(result_split[:-1], self.__available_parameters, tag_number)
                    if self.parameters:
                        return True
                except Exception:
                    pass

    async def flush(self):
        await self.__serial_data.flush()

    def __send_command(self, command):
        command = command + '\r'
        self.__serial.write(command)

    async def __get_sensor_response(self, manta_command):
        sensor_response = MantaResponse(manta_command)
        response_timeout = obs.millis()
        self.__send_command(manta_command)
        while True:
            returned_response = await self.__serial_data.update()
            if returned_response:
                try:
                    sensor_response.value = returned_response[0].decode()
                    sensor_response.value = sensor_response.value.replace('\r\n', '')
                    return sensor_response
                except Exception:
                    pass
            if obs.elapsed_millis(response_timeout) > 5000:
                return sensor_response
            await uasyncio.sleep_ms(100)
        return sensor_response

