import omc048
import obs
import config
import uasyncio
from read_line import Read_line
from nmea import parser
from sensor_drv import SensorDrv
from util import parse_int, parse_float, datetime_update, utc_synchronisation

PARAM = 0
VALUE = 1

class NavilockMd6(SensorDrv):
  
  def __init__(self, serial, conf):
    SensorDrv.__init__(self)
    self.__sensor_data_valid = False
    self.__data = {}
    self.__params = {}
    self.__serial_data = Read_line(serial)
    self.__time_deviation_in_sec = 2.0
    self.__utc_sync = config.get_config_or_default(conf, 'utc_time_sync', False)
    self.__utc_offset = omc048.RTC().utc_offset()
    self.__syslog = obs.system_log('GPS')
    self.__nmea_parser = parser()
    if self.__utc_sync:
      self.__syslog.info('UTC automatic time synchronisation enabled on {}'.format(conf['id']))
 
  async def read_values(self, detect_params=False):
    response_timeout = obs.millis()
    self.__sensor_data_valid = False
    self.__data = {}
    while True:
      data = await self.__serial_data.update()
      if data:
        for line in data:
          nmea_data_tagged = self.decode(line)
          if nmea_data_tagged and self._update(nmea_data_tagged, detect_params):
            return True
      if obs.elapsed_millis(response_timeout) > 3000:
          break
      await uasyncio.sleep_ms(1)

  def gga_decode(self, data, detect_params):
    lat = self.convert_lat(data[2][VALUE], data[3][VALUE])
    lon = self.convert_lon(data[4][VALUE], data[5][VALUE])
    #Only use the GPS data when quality index is 1 
    if data[6][VALUE] == '1':
      self.__sensor_data_valid = True
    self.__data['GGA'] = [lat,lon,data[6][VALUE],data[7][VALUE],data[8][VALUE],data[9][VALUE],data[10][VALUE],data[11][VALUE],data[12][VALUE]]
    if detect_params:
      self.__params['GGA'] = [data[2][PARAM],data[4][PARAM],data[6][PARAM],data[7][PARAM],data[8][PARAM],data[9][PARAM],data[10][PARAM],data[11][PARAM],data[12][PARAM]]

  def vtg_decode(self, data, detect_params):
    self.__data['VTG'] = [data[1][VALUE],data[2][VALUE],data[3][VALUE],data[4][VALUE],data[5][VALUE]]
    if detect_params:
      self.__params['VTG'] = [data[1][PARAM],data[2][PARAM],data[3][PARAM],data[4][PARAM],data[5][PARAM]]

  def rmc_decode(self, data, detect_params):
    if self.__utc_sync and data[1][VALUE] and data[9][VALUE]:
      utc_synchronisation(datetime_update(data[1][VALUE], data[9][VALUE]), self.__utc_offset[0], self.__utc_offset[1], self.__time_deviation_in_sec)

  def txt_decode(self, data, detect_params):
    #Do nothing with the TXT data
    pass

  _nmea_senid = {
        "GGA" : gga_decode,
        "RMC" : rmc_decode,
        "VTG" : vtg_decode,
        "TXT" : txt_decode
	}

  def _update(self, data, detect_params):
    if not data:
      return False
    try:
      decoder = self._nmea_senid[data[0][VALUE][2:]]
      decoder(self, data, detect_params)
      if len(self.__data) == 2 and self.__sensor_data_valid:
        self.data = self.__data['GGA'] + self.__data['VTG']
        self.__sensor_data_valid = False
        self.__data = {}
        return True
      elif len(self.__params) == 2 and detect_params:
        return True
    except KeyError:
      self.__syslog.warn('Unsupported decoded format: {}'.format(data))
    return False

  #Latitude is converted to DD (decimal degrees) notation due to omc045 log format    
  def convert_lat(self, lat, dir_lat):
    if lat:
      lat_degs = parse_int(lat[0:2])
      lat_mins = parse_float(lat[2:])
      decimal_degrees = lat_degs + (lat_mins / 60)
      return -decimal_degrees if dir_lat == 'S' else decimal_degrees

  #Longitude is converted to DD (decimal degrees) notation due to omc045 log format   
  def convert_lon(self, lon, dir_lon):
    if lon:
      lon_degs = parse_int(lon[0:3])
      lon_mins = parse_float(lon[3:])

      if lon_degs and lon_mins:
        decimal_degrees = lon_degs + (lon_mins / 60)
        return -decimal_degrees if dir_lon == 'W' else decimal_degrees
  
  async def detect_parameters(self, tag_number = ''):
    if await self.read_values(True):
      params = self.__params['GGA'] + self.__params['VTG']
      for param in params:
        if param[2] == 'LAT' or param[2] == 'LON':
          self.parameters.append((param[0], param[1], param[2]))  
        else:
          self.parameters.append((param[0], param[1], param[2] + tag_number))
      return True

  async def flush(self):
    await self.__serial_data.flush()

  #Filter all empty sentence fields ('') data to be converted to 'None'
  def _filter_empty_fields(self, msg):
    for x in msg:
      if x == '':
        msg[msg.index(x)] = None
    return msg

###   GGA   ##################################################################
  def _gga(self, msg):
    if len(msg) < 14:
        return None
    return [((), msg[0]),
            (('UTC', '', 'TIME'), msg[1]),
            (('Latitude', 'deg', 'LAT'), msg[2]),
            (('Dir of lat', '', 'DLAT'), msg[3]),
            (('Longitude', 'deg', 'LON'), msg[4]),
            (('Dir of lon', '', 'DLON'), msg[5]),
            (('Quality', '', 'GPSQ'), msg[6]),
            (('Satellites', '', 'SAT'), msg[7]),
            (('Hdop', 'm', 'HDOP'), msg[8]),
            (('Altitude', 'm', 'ALT'), msg[9]),
            (('Geoid sep', '', 'GEO'), msg[11]),
            (('Age', 'min', 'AGE'), msg[13]),
            (('Station', '', 'STAT'), msg[14])]

###   VTG   ##################################################################
  def _vtg(self, msg):
    if len(msg) < 9:
        return None
    return [((), msg[0]),
            (('Course true', 'deg', 'CRST'), msg[1]),
            (('Course magnetic', 'deg', 'CRSM'), msg[3]),
            (('Speed knots', 'knots', 'SPDK'), msg[5]),
            (('Speed kmh', 'km/h', 'SPDM'), msg[7]),
            (('Mode', '', 'MOD'), msg[9])]

###   RMC   ##################################################################
  def _rmc(self, msg):
    if len(msg) < 12:
        return None
    return [((), msg[0]),
            (('UTC', '', 'TIME'), msg[1]),
            (('Status', '', 'STS'), msg[2]),
            (('Latitude', 'deg', 'LAT'), msg[3]),
            (('Dir of lat', '', 'DLAT'), msg[4]),
            (('Longitude', 'deg', 'LON'), msg[5]),
            (('Dir of lon', '', 'DLON'), msg[6]),
            (('Speed knots', 'knots', 'SPDK'), msg[7]),
            (('Course true', 'deg', 'CRST'), msg[8]),
            (('Date', '', 'DATE'), msg[9]),
            (('Variation', 'deg', 'VAR'), msg[10]),
            (('Mode', '', 'MOD'), msg[11]),
            (('Nav status', '', 'NGST'), msg[12])]

###   TXT   ##################################################################
  def _txt(self, msg):
    if len(msg) < 4:
        return None
    return (((), msg[0]),
            (('Number of sentences', '', 'NOS'), msg[1]),
            (('Sentence number', '', 'SENN'), msg[2]),
            (('Text id', '', 'TXID'), msg[3]),
            (('Message', '', 'MSSG'), msg[4]))

  def decode(self, data, checksum=True):
    msg = self.__nmea_parser.parse(data, checksum)
    if msg:
      if msg[0][0] == 'P':
        decoder = self.decoders.get(msg[0])
      else:
        decoder = self.decoders.get(msg[0][2:])
      if decoder is not None:
        return decoder(self,self._filter_empty_fields(msg))
  
###   SUPPORTED DECODERS   ####################################################
  decoders = {  'TXT': _txt, 
                'GGA': _gga,
                'VTG': _vtg,
                'RMC': _rmc}


