#Note: After discussion with Rene on why the ysi_6_series is operating in run mode while the EXO is polled:
#The ysi-6-series is slow to respond on the data command in case the 'sample and hold' function is enabled (default).
#If the 'sample and hold' is disabled the sensor does not respond to wipe commands, meaning there is no control over the wiper.
#By using the 'run' mode the wiper is controlled by the sensor itself, meaning that data requests can be delayed in case the sensor is always powered.
#Therefore it is recommended to always use the ysi-6-series using 'supply_port_always_on: False'

import re
import obs
import omc048
import config
import uasyncio
from read_line import Read_line
from sensor_drv import SensorDrv
from util import parse_int, strlist_to_intlist, lookup_parameters

class Ysi6Series(SensorDrv):
    __syslog = obs.system_log('YSI-6')
    
    def __init__(self, serial, conf):
        SensorDrv.__init__(self)
        self.__available_parameters = {    51:     ("Date d/m/y"        ,"d/m/y"       ,"Date1" ), 
                                            52:     ("Date m/d/y"        ,"m/d/y"       ,"Date2" ), 
                                            53:     ("Date y/m/d"        ,"y/m/d"       ,"Date3" ), 
                                            153:    ("Date"              ,"UNKNOWN"     ,"DATE" ), 
                                            54:     ("Time hh:mm:ss"     ,"hh:mm:ss"    ,"Time1" ), 
                                            1:      ("Temp C"	         ,"C"           ,"TMPC" ), 
                                            2:      ("Temp F"	         ,"F"           ,"TMPF" ), 
                                            3:      ("Temp K"	         ,"K"           ,"TMPK" ), 
                                            6:      ("SpCond ms/cm"      ,"mS/cm"       ,"SPCm"  ), 
                                            7:      ("SpCond uS/cm"      ,"uS/cm"       ,"SPCu"  ), 
                                            4:      ("Cond mS/cm"	     ,"mS/cm"       ,"CONm"  ), 
                                            5:      ("Cond uS/cm"        ,"uS/cm"       ,"CONu"  ), 
                                             9:     ("Resist MOhm*cm"    ,"MOhm*cm"     ,"RESm" ), 
                                             8:     ("Resist KOhm*cm"    ,"KOhm*cm"     ,"RESk" ), 
                                            94:     ("Resist Ohm*cm"     ,"Ohm*cm"      ,"RESo" ), 
                                            10:     ("TDS g/L"           ,"g/L"         ,"TDSg"  ), 
                                            95:     ("TDS Kg/L"          ,"Kg/L"        ,"TDSk" ), 
                                            12:     ("Sal ppt"           ,"ppt"         ,"SALp" ), 
                                            14:     ("DOsat %"           ,"%"           ,"DOs" ), 
                                            200:    ("DOsat %Local"      ,"%"           ,"DOl" ), 
                                            15:     ("DO mg/L"           ,"mg/L"        ,"DOm" ), 
                                            96:     ("DOchrg"            ,"chrg"        ,"DOc" ), 
                                            209:    ("Cl2 mg/L"          ,"mg/L"        ,"CLm" ), 
                                            210:    ("Cl2chrg"           ,"chrg"        ,"CLc" ), 
                                            20:     ("Press psia"        ,"psia"        ,"PSIa" ), 
                                            104:    ("Press psir"        ,"psir"        ,"PSIr" ), 
                                            21:     ("Press psig"        ,"psig"        ,"PSIg" ), 
                                            111:    ("Press psi"         ,"psi"         ,"PSI" ), 
                                            22:     ("Depth meters"      ,"m"           ,"DEPm" ), 
                                            23:     ("Depth feet"        ,"f"           ,"DEPf" ), 
                                            118:    ("Flow ft3/sec"      ,"ft3/sec"     ,"FLOW" ), 
                                            166:    ("Flow ft3/min"      ,"ft3/min"     ,"FLOW" ), 
                                            167:    ("Flow ft3/hour"     ,"ft3/hour"    ,"FLOW" ), 
                                            168:    ("Flow ft3/day"      ,"ft3/day"     ,"FLOW" ), 
                                            164:    ("Flow gal/sec"      ,"gal/sec"     ,"FLOW" ), 
                                            119:    ("Flow gal/min"      ,"gal/min"     ,"FLOW" ), 
                                            165:    ("Flow gal/hour"     ,"gal/hour"    ,"FLOW" ), 
                                            120:    ("Flow Mgal/day"     ,"Mgal/day"    ,"FLOW" ), 
                                            121:    ("Flow m3/sec"       ,"m3/sec"      ,"FLOW" ), 
                                            169:    ("Flow m3/min"       ,"m3/min"      ,"FLOW" ), 
                                            170:    ("Flow m3/hour"      ,"m3/hour"     ,"FLOW" ), 
                                            171:    ("Flow m3/day"       ,"m3/day"      ,"FLOW" ), 
                                            122:    ("Flow L/s"          ,"L/s"         ,"FLOW" ), 
                                            172:    ("Flow AF/day"       ,"AF/day"      ,"FLOW" ), 
                                            123:    ("Volume ft3"        ,"ft3"         ,"VOL" ), 
                                            124:    ("Volume gal"        ,"gal"         ,"VOL" ), 
                                            173:    ("Volume Mgal"       ,"Mgal"        ,"VOL" ), 
                                            125:    ("Volume m3"         ,"m3"          ,"VOL" ), 
                                            126:    ("Volume L"          ,"L"           ,"VOL" ), 
                                            174:    ("Volume acre*ft"    ,"acre*ft"     ,"VOL" ), 
                                            18:     ("pH"                ,"pH"          ,"PH" ), 
                                            17:     ("pH mV"             ,"mV"          ,"PHv" ), 
                                            19:     ("Orp mV"            ,"mV"          ,"ORPv" ), 
                                            48:     ("NH4+ N mg/L"       ,"mg/L"        ,"NH4m" ), 
                                            108:    ("NH4+ N mV"         ,"mV"          ,"NH4v" ), 
                                            47:     ("NH3 N mg/L"        ,"mg/L"        ,"NH3m" ), 
                                            106:    ("NO3- N mg/L"       ,"mg/L"        ,"NO3m" ), 
                                            101:    ("NO3- N mV"         ,"mV"          ,"NO3v" ), 
                                            112:    ("Cl- mg/L"          ,"mg/L"        ,"CLm" ), 
                                            145:    ("Cl- mV"            ,"mV"          ,"CLv" ), 
                                            201:    ("PAR1"              ,""            ,"PAR1" ), 
                                            202:    ("PAR2"              ,""            ,"PAR2" ), 
                                            37:     ("Turbid NTU"        ,"NTU"         ,"TURn" ), 
                                            203:    ("Turbid+ NTU"       ,"NTU"         ,"TURPn" ), 
                                            193:    ("Chl ug/L"          ,"ug/L"        ,"CHLu" ), 
                                            194:    ("Chl RFU"           ,"RFU"         ,"CHLf" ), 
                                            204:    ("Rhodamine ug/L"    ,"ug/L"        ,"RHOu" ), 
                                            211:    ("ODOsat %"          ,"%"           ,"ODOs" ), 
                                            214:    ("ODOsat %Local"     ,"%"           ,"ODOl" ), 
                                            212:    ("ODO mg/L"          ,"mg/L"        ,"ODOm" ),
                                            215:    ("BGA-PC cells/mL"   ,"cells/mL"    ,"BGACc" ), 
                                            216:    ("BGA-PC RFU"        ,"RFU"         ,"BGACf" ), 
                                            217:    ("BGA-PE cells/mL"   ,"cells/mL"    ,"BGAEc" ), 
                                            218:    ("BGA-PE RFU"        ,"RFU"         ,"BGAEf" ), 
                                            98:     ("Gnd Hz"            ,"Hz"          ,"GND" ), 
                                            99:     ("Scale Hz"          ,"Hz"          ,"SCL" ), 
                                            100:    ("Prescmp"           ,""            ,"PRES" ), 
                                            32:     ("Density kg/m3"     ,"kg/m3"       ,"DNSTY" ), 
                                            28:     ("Battery volts"     ,"v"           ,"BTV" )}
        self.startup_time = 30
        self.__para_re = re.compile(b'^([0-9* ]*)\r\n$')
        self.__data_re = re.compile(b'^(( )*([!-~]+))+\r\n$')
        self.__data_multiple_spaces = re.compile(' +')
        self.__data_single_space = re.compile('^ *')
        self.__serial_data = Read_line(serial)
        self.__serial = serial
        self.__parameter_size = 0

    async def read_values(self):  
        #YSI sends data automatically, better to always flush old data
        await self.flush()
        response = await self.__get_sensor_response()
        if response.value and self.__check_data_response(response.value):
            sensor_response = response.value.decode()
            sensor_response = sensor_response.replace('\r\n','')
            sensor_response = self.__data_multiple_spaces.sub(' ', sensor_response)
            sensor_response = self.__data_single_space.sub('', sensor_response)
            response_split = sensor_response.split(' ')
            if len(response_split) == self.__parameter_size:
                self.data = response_split
                return True

    def __send_command(self, command):
        command = command + '\r'
        self.__serial.write(command)

    async def __get_sensor_response(self, ysi_command=None):
        sensor_response = ysi_response()
        response_timeout = obs.millis()
        if ysi_command:
            self.__send_command(ysi_command)
        while True:
            returned_response = await self.__serial_data.update()
            if returned_response:
                if ysi_command:
                    if self.assemble_response(sensor_response, returned_response):
                        return sensor_response
                else:
                    sensor_response.value = returned_response[0]
                    return sensor_response

            if obs.elapsed_millis(response_timeout) > 1000:
               return ysi_response()
            await uasyncio.sleep_ms(100)
        return ysi_response()

    def assemble_response(self, response_class, response_data):
        if len(response_data) == 1 and not response_class.command:
            response_class.command = response_data[0]
        elif len(response_data) == 1 and response_class.command:
            response_class.value = response_data[0]
            return True
        else:
            response_class.command = response_data[0]
            response_class.value = response_data[1]
            return True

    async def __retreive_parameters(self):
        # escape = \x1B
        #YSI sends data automatically, better to always flush old data
        await self.flush()
        await self.__get_sensor_response('\x1B')
        response = await self.__get_sensor_response('para')
        if response.value and self.__check_parameter_response(response.value):
            decoded_response = response.value.decode()
            response_split = decoded_response.split(' ')
            response_int = strlist_to_intlist(response_split)
            self.__parameter_size = len(response_int)
            self.__send_command('run')
            return response_int
        return None

    def __check_parameter_response(self, sensor_response):
        if not self.__para_re.match(sensor_response):
            self.__syslog.warn("incorrect parameter format received: {}".format(sensor_response))
            return False
        return True
 
    def __check_data_response(self, sensor_response):
        if not self.__data_re.match(sensor_response):
            self.__syslog.warn("incorrect data format received: {}".format(sensor_response))
            return False
        return True

    async def detect_parameters(self, tag_number = ''):
        resp = await self.__retreive_parameters()
        if resp:
            self.parameters = lookup_parameters(resp, self.__available_parameters, tag_number)
            return True
        return False
        
    async def flush(self):
        await self.__serial_data.flush()

class ysi_response:
    command = ''
    value = ''

